# (c) Nick Polyak 2023
# License: MIT License (https://opensource.org/licenses/MIT)


# import the client stubs (service_pb2 contains messages, 
# service_pb2_grpc contains RPCs)
import RelayService_pb2 as relay_service
from google.protobuf import message
from google.protobuf import any_pb2
from RelayClientBase import RelayClientBase

class BroadcastingRelayClient(RelayClientBase):
    def __init__(self, hostname:str, port:int):
        super(BroadcastingRelayClient, self).__init__(hostname, port, False)

    def broadcast(self, msg:relay_service.FullMsg) -> relay_service.ShortMsg:
        return self._client_stub.PublishTopic(msg)

    def broadcast_object(self, item:message.Message, topic_name:str, topic_number:int) -> relay_service.ShortMsg:
        a = any_pb2.Any()
        a.Pack(item)
        metadata = self.create_short_msg(topic_name="PersonTopic", topic_number=1)
        msg = relay_service.FullMsg(metadata = metadata, message = a)
        return self.broadcast(msg)



# (c) Nick Polyak 2023
# License: MIT License (https://opensource.org/licenses/MIT)

# import python packages
import asyncio
import concurrent.futures
import grpc
import datetime

# import the client stubs (service_pb2 contains messages, 
# service_pb2_grpc contains RPCs)
import RelayService_pb2 as relay_service
import RelayService_pb2_grpc as relay_service_grpc
from google.protobuf import type_pb2 as proto_types
from google.protobuf import timestamp_pb2

class RelayClientBase:
    def __init__(self, hostname:str, port:int, is_async:bool=True):
        self._is_async = is_async
        self._hostname = hostname
        self._port = port;
        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
        self._client_stub = None
    
    def __create_channel__(self) -> relay_service_grpc.RelayServiceStub:
        hostport = self._hostname + ":" + str(self._port)
        if self._is_async:
            self._channel = grpc.aio.insecure_channel(hostport)
        else:
            self._channel = grpc.insecure_channel(hostport)
        stub = relay_service_grpc.RelayServiceStub(self._channel)
        return stub

    def connect_if_needed(self):
        if (self._client_stub == None):
            self._client_stub = self.__create_channel__()

    def create_short_msg(self, topic_name:str, topic_number:int) -> relay_service.ShortMsg:
        current_utc_date_time = datetime.datetime.now(datetime.timezone.utc)
        current_utc_timestamp = timestamp_pb2.Timestamp()
        current_utc_timestamp.FromDatetime(current_utc_date_time)

        topic_enum_val = proto_types.EnumValue(name=topic_name, number=topic_number)
        msg = relay_service.ShortMsg(topic=topic_enum_val, msgSentTime=current_utc_timestamp)

        return msg








# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: RelayService.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from google.protobuf import type_pb2 as google_dot_protobuf_dot_type__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12RelayService.proto\x12\x19nP.grpc.relayServiceProto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x19google/protobuf/any.proto\x1a\x1agoogle/protobuf/type.proto\"|\n\x08ShortMsg\x12/\n\x0bmsgSentTime\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12)\n\x05topic\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.EnumValue\x12\x14\n\x0c\x65rrorMessage\x18\x04 \x01(\t\"g\n\x07\x46ullMsg\x12\x35\n\x08metadata\x18\x01 \x01(\x0b\x32#.nP.grpc.relayServiceProto.ShortMsg\x12%\n\x07message\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any2\xc8\x01\n\x0cRelayService\x12]\n\x0eGetTopicStream\x12#.nP.grpc.relayServiceProto.ShortMsg\x1a\".nP.grpc.relayServiceProto.FullMsg\"\x00\x30\x01\x12Y\n\x0cPublishTopic\x12\".nP.grpc.relayServiceProto.FullMsg\x1a#.nP.grpc.relayServiceProto.ShortMsg\"\x00\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'RelayService_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SHORTMSG._serialized_start=137
  _SHORTMSG._serialized_end=261
  _FULLMSG._serialized_start=263
  _FULLMSG._serialized_end=366
  _RELAYSERVICE._serialized_start=369
  _RELAYSERVICE._serialized_end=569
# @@protoc_insertion_point(module_scope)



# (c) Nick Polyak 2023
# License: MIT License (https://opensource.org/licenses/MIT)

# import python packages
import asyncio
from re import T
import grpc
import reactivex
from typing import TypeVar

# import the client stubs (service_pb2 contains messages, 
# service_pb2_grpc contains RPCs)
import RelayService_pb2 as relay_service
from RelayClientBase import RelayClientBase
from google.protobuf import message
from google.protobuf import any_pb2
import reactivex.operators as ops

class ObservingRelayClient(RelayClientBase):
    async def __observe_topic_async_impl__(self, t_name:str, t_number:int) -> None:
        self.connect_if_needed()

        requestMessage = self.create_short_msg(topic_name=t_name, topic_number=t_number)

        self._stream = self._client_stub.GetTopicStream(requestMessage)

        try:
            async for msg in self._stream:
                self._observable.on_next(msg)
        except grpc.RpcError as rpc_error:
            if (rpc_error.code() == grpc.StatusCode.CANCELLED):
                self._observable.on_completed()
            else:
                self._observable.on_error(rpc_error)

    def __run_observe_topic_impl__(self, topic_name:str, topic_number:int):
        asyncio.run((self.__observe_topic_async_impl__(topic_name, topic_number)))

    def observe_topic(self, topic_name:str, topic_number:int) -> None:
        self._observable = reactivex.Subject[relay_service.FullMsg]()

        self._executor.submit(self.__run_observe_topic_impl__, topic_name, topic_number)

    def cancel(self):
        self._stream.cancel()

    def get_observable(self) -> reactivex.Observable[relay_service.FullMsg]:
        return self._observable;

    T = TypeVar("T", bound=message.Message)
    def get_concrete_observable(self, create_T_method) -> reactivex.Observable[T]:
        def get_T(msg:relay_service.FullMsg) -> T:
            t = create_T_method();
            msg.message.Unpack(t);
            return t;
        return self._observable.pipe(ops.map(lambda t : get_T(t)))